import std.string;
import std.regexp;
import std.file;
import std.stdio;

import d2substs;


string VERSION = "0.1";


struct TranslationInfo
{
    string[] locations;
    string translation;
    string toString() { return translation; }
}


TranslationInfo[string] extractTranslationStrings(string[] files)
{
    RegExp i18n_string = new RegExp(`.*_\("(.*)"\).*`);
    TranslationInfo[string] TRANSLATIONS;

    int nextIdx = 0;
    foreach(f; files)
        foreach(uint i, line; splitlines(cast(string)std.file.read(f)))
            foreach(m; i18n_string.search(line))
            {
                string key = m.match(1);
                
                if((key in TRANSLATIONS) == null)
                    TRANSLATIONS[key] = TranslationInfo();
                
                TRANSLATIONS[key].locations ~= [format("%s:%d", f, i+1)];
            }
    
    foreach(key; TRANSLATIONS.keys)
        TRANSLATIONS[key].locations.sort;
    
    return TRANSLATIONS;
}


string formatToPoOrderedByFile(TranslationInfo[string] _messages)
{
    string[string] orderList;
    string poFormattedString;
    
    foreach(key; _messages.keys)
        orderList[_messages[key].locations[0]] ~= key;

    foreach(orderKey; orderList.keys.sort)
        foreach(key; _messages.keys)
            if(_messages[key].locations[0] == orderKey)
                poFormattedString ~= formatNodeToPo(key, _messages[key]);

    return poFormattedString;
}


string formatToPo(TranslationInfo[string] _translations)
{
    string transFileContent;
    foreach(key; _translations.keys)
        transFileContent ~= formatNodeToPo(key, _translations[key]);
    return transFileContent;
}


string formatNodeToPo(string toTranslate, TranslationInfo transInfo)
{
    string locations;
    foreach(location; transInfo.locations)
        locations ~= format("#: %s\n", location);
    return format("%s\nmsgid \"%s\"\nmsgstr \"%s\"\n\n", locations[0..length-1], escape(toTranslate), escape(transInfo.translation));
}



string escape(string _str)
{
    return _str.replace("%", "%%");
}



string unescape(string _str)
{
    return _str.replace("%%", "%");
}


class PoFileParser
{
    this(string _translationsFilePath)
    {
        if( (_translationsFilePath == null) || (!std.file.exists(_translationsFilePath)) )
        {
            string msg = "constructor's argument 'translationsFilePath' can't "
                         "be 'null' and must be a valid and existent file!";
            throw new PoFileParserException(msg);
        }
        translationsFilePath = _translationsFilePath;
    }
    
    TranslationInfo[string] getParsed()
    {
        if(MESSAGES == null) parse();
        return MESSAGES;
    }
    
    
    void parse()
    {
        RegExp idLine = new RegExp(`msgid\s*"(.*)"\s*$`);
        RegExp strLine = new RegExp(`msgstr\s*"(.*)"\s*$`);
        RegExp textLine = new RegExp(`"(.*)"`);
    
        string lastOpen = null;
        string key, value;
        
        foreach(uint i, line; readFileIntoLines(translationsFilePath))
        {
            if (isComment(line) || isBlank(line))
            {
                debug
                {
                    std.stdio.fwritefln(stderr, format("PoFileParser DEBUG: ignored #%d", i+1));
                }
                continue;
            }

            else if (string[] m = idLine.match(line))
            {
                debug{std.stdio.fwritefln(stderr, format("PoFileParser DEBUG: #%d id  '%s'", i+1, m[1].replace("%", "%%")));}
                if (lastOpen != "msgid")
                {
                    if (key != null && value != null) messagePair(key, value);
                    key = m[1];
                    lastOpen = "msgid";
                }
                else
                {
                    string msg = "%s(%d): Error in file format!!! Two consecutive 'msgid' lines";
                    throw new PoFileParserException(format(msg, translationsFilePath, i+1));
                }
            }

            else if (string[] m = strLine.match(line))
            {
                debug {std.stdio.fwritefln(stderr, format("PoFileParser DEBUG: #%d str '%s'", i+1, m[1].replace("%", "%%")));}
                if (lastOpen != "msgstr")
                {
                    value = m[1];
                    lastOpen = "msgstr";
                }
                else
                {
                    string msg = "%s(%d): Error in file format!!! Two consecutive 'msgstr' lines";
                    throw new PoFileParserException(format(msg, translationsFilePath, i+1));
                }
            }

            else
            {
                if (lastOpen == "msgid") key ~= textLine.match(line)[1];
                if (lastOpen == "msgstr") value ~= textLine.match(line)[1];
            }
        }
        messagePair(key, value);
    }


private:
    string translationsFilePath;
    TranslationInfo[string] MESSAGES = null;

 
    string[] readFileIntoLines(string _filePath)
    {
        string content = cast(string)std.file.read(_filePath);
        return std.string.splitlines(content);
    }


    void messagePair(string _key, string _value)
    {
        _key = _key.replace("\\n", "\n");
        _value = _value.replace("\\n", "\n");
        
        if((_key in MESSAGES) == null)
                    MESSAGES[_key] = TranslationInfo();
        
        MESSAGES[_key].translation = _value;
    }


    bool isComment(string _string) { return startsWith(_string, "#"); }


    bool isBlank(string _string) { return (strip(_string).length == 0); }

    
    class PoFileParserException : Exception { this(string _msg) { super(_msg); } }
}

